"""Implementation of FBA Problem class.

Peter Schubert, HHU Duesseldorf, June 2022
"""

import numpy as np

from xbanalysis.problems.lp_problem import LpProblem


class FbaProblem:

    def __init__(self, xba_model):
        """Initialize FbaProblem from a xba_model.

        raise an exceptions, if fba parameters are not configured in the xba_model

        FBA mass balanced constraints: S * v = 0
        - rows of stoichiometric matrix S:
          - all metabolites (incl. boundary/constant metabolites),
          - i.e. excluding enzymes
        - columns of stoichiometric matrix S:
          - all reactions not containing enzymes as reactants or products
          - i.e. exlcluding protein synthesis, protein degradation, enzyme transitions
          - reversible reactions are not getting split into forward and backward reaction (for now)

        FBA flux bounds:
        - lower and upper bounds for reactions extracted from xba_model
        - getter and setter methods

        FBA objective: c.T * v:
        - extracted from xba_model (fbcObjective), including direction
        - getter and setter methods

        :param xba_model: xba model with all required FBA parameters
        :type xba_model: XbaModel
        """
        self.xba_model = xba_model

        # identify reaction excluding protein synthesis and protein degradation
        self.mr_rids = [r.id for r in xba_model.reactions.values()
                        if r.sboterm.is_in_branch('SBO:0000205') is False and
                        r.sboterm.is_in_branch('SBO:0000179') is False]
        self.mrid2idx = {mrid: idx for idx, mrid in enumerate(self.mr_rids)}
        self.reversible = np.array([self.xba_model.reactions[rid].reversible for rid in self.mr_rids])

        sids = set()
        for rid in self.mr_rids:
            for sid in xba_model.reactions[rid].reactants:
                sids.add(sid)
            for sid in xba_model.reactions[rid].products:
                sids.add(sid)
        self.sids = list(sids)

        self.s_mat = xba_model.get_stoic_matrix(self.sids, self.mr_rids)

        self.is_fba_model = self.check_fba_model(xba_model)
        if self.is_fba_model is False:
            print('Error: not a FBA model - missing parameters')
            return

        self.flux_bounds = np.array([[xba_model.reactions[rid].fbc_lb for rid in self.mr_rids],
                                     [xba_model.reactions[rid].fbc_ub for rid in self.mr_rids]])
        self.obj_coefs = None
        for oid, fbc_obj in xba_model.fbc_objectives.items():
            if fbc_obj.active is True:
                self.obj_id = fbc_obj.id
                self.obj_dir = fbc_obj.direction
                self.set_obj_coefs(fbc_obj.coefficiants)
                break

    def check_fba_model(self, xba_model):
        """Check if fba related data available in the XBA model

        :param xba_model: Xba model with data extracted from SBML file
        :type xba_model: XbaModel
        :return: indicator if model contains data required for FBA problem formulation
        :rtype: bool
        """
        is_fba_model = True

        r_check = {'flux lower bound': 'fbc_lb', 'flux upper bound': 'fbc_ub'}
        for check, param in r_check.items():
            missing = [rid for rid in self.mr_rids if hasattr(xba_model.reactions[rid], param) is False]
            if len(missing) > 0:
                print(f'missing {check} in:', missing)
                is_fba_model = False

        if hasattr(xba_model, 'fbc_objectives') is False:
            print('missing flux objectives')
            is_fba_model = False
        else:
            active_objs = [oid for oid in xba_model.fbc_objectives if xba_model.fbc_objectives[oid].active]
            if len(active_objs) == 0:
                print('missing an active flux objective')
                is_fba_model = False

        return is_fba_model

    def combine_fwd_bwd(self, vector):
        """combine forward and backward flux information.

        :param vector: contains data on all reactions + backward information for reversible reactions
        :type vector: 1D ndarray of shape (len(mr_rids) + sum(reversible))
        :return: combined flux information
        :rtype: 1D ndarray of shape (len(mr_rids))
        """
        n_cols = len(self.mr_rids)
        combined = vector[:n_cols]
        combined[self.reversible] -= vector[n_cols:]
        return combined

    def create_fba_lp(self):
        """create Linear Programming Problem for FBA with split reactions

        Splitting reactions in forward/backward directions reduces high values during FBA
        Splitting required for pFBA (minimize absolute flux values)

        :return: LpProblem configured for FBA
        :rtype: LpProblem
        """
        if self.is_fba_model is False:
            return None

        lp = LpProblem()
        # self.reversible = np.array([self.xba_model.reactions[rid].reversible for rid in self.mr_rids])
        fwd_bwd_s_mat = np.hstack((self.s_mat, -self.s_mat[:, self.reversible]))
        fwd_bwd_obj_coefs = np.hstack((self.obj_coefs, -self.obj_coefs[self.reversible]))
        fwd_bwd_bnds = np.hstack((np.fmax(self.flux_bounds, 0),
                                  np.vstack((np.fmax(-self.flux_bounds[1, self.reversible], 0),
                                             np.fmax(-self.flux_bounds[0, self.reversible], 0)))))
        row_bnds = np.zeros((fwd_bwd_s_mat.shape[0], 2))
        lp.configure(fwd_bwd_obj_coefs, fwd_bwd_s_mat, row_bnds, fwd_bwd_bnds, self.obj_dir)
        return lp

    def fba_optimize(self):
        """FBA optimization of current problem

        :return: results information from the optimization
        :rtype: dict
        """
        lp = self.create_fba_lp()
        if lp is None:
            return {'success': False, 'message': 'not an FBA model'}
        res = lp.solve()
        del lp
        res['x'] = self.combine_fwd_bwd(res['x'])
        res['reduced_costs'] = self.combine_fwd_bwd(res['reduced_costs'])
        return res

    def pfba_optimize(self, fract_of_optimum=1.0):
        """ pfba analysis

        Reactions need to be split in forward/backard (to minimize absolute flux value)

        :param fract_of_optimum: value between 0.0 and 1.0.
        :type fract_of_optimum: float (default 1.0)
        :return:
        """
        lp = self.create_fba_lp()
        if lp is None:
            return {'success': False, 'message': 'not an FBA model'}
        res = lp.solve(short_result=False)
        if res['success']:
            # fix FBA objective as constraint
            fwd_bwd_obj_coefs = np.hstack((self.obj_coefs, -self.obj_coefs[self.reversible]))
            if self.obj_dir == 'maximize':
                row_bds = np.array([res['fun'] * fract_of_optimum, np.nan])
            else:
                row_bds = np.array([np.nan, res['fun'] / fract_of_optimum])
            lp.add_constraint(fwd_bwd_obj_coefs, row_bds)
            # new LP objective is to minimize sum of fluxes
            n_cols = len(self.mr_rids)
            lp.set_obj_coefs(np.ones(n_cols + sum(self.reversible)))
            lp.set_obj_dir('minimize')
            res = lp.solve()
            res['x'] = self.combine_fwd_bwd(res['x'])
            res['reduced_costs'] = self.combine_fwd_bwd(res['reduced_costs'])
        del lp
        return res

    def fva_optimize(self, rids=None, fract_of_optimum=1.0):
        """FVA - flux variability analysis for all or selected reactions.

        :param rids: selected reaction ids to run FVA (alternatively, all)
        :type rids: list of strings (default: None - all reactions)
        :param fract_of_optimum: value between 0.0 and 1.0.
        :type fract_of_optimum: float (default 1.0)
        :return: Flux ranges and objective values
        :rtype: pandas DataFrame
        """
        lp = self.create_fba_lp()
        if lp is None:
            return {'success': False, 'message': 'not an FBA model'}
        res = lp.solve(short_result=True)

        if rids is None:
            rids = self.mr_rids
        flux_ranges = np.zeros((4, len(rids)))
        if res['success']:
            # fix FBA objective as constraint
            fwd_bwd_obj_coefs = np.hstack((self.obj_coefs, -self.obj_coefs[self.reversible]))
            if self.obj_dir == 'maximize':
                row_bds = np.array([res['fun'] * fract_of_optimum, np.nan])
            else:
                row_bds = np.array([np.nan, res['fun'] / fract_of_optimum])
            fba_constr_row = lp.add_constraint(fwd_bwd_obj_coefs, row_bds)

            n_cols = len(self.mr_rids)
            for idx, rid in enumerate(rids):
                # select reaction maximize/minimize flux
                new_obj_coefs = np.zeros(n_cols)
                new_obj_coefs[self.mrid2idx[rid]] = 1
                new_fwd_bwd_obj_coefs = np.hstack((new_obj_coefs, -new_obj_coefs[self.reversible]))
                lp.set_obj_coefs(new_fwd_bwd_obj_coefs)

                lp.set_obj_dir('minimize')
                res = lp.solve(short_result=True)
                if res['success']:
                    flux_ranges[0, idx] = res['fun']
                    flux_ranges[2, idx] = lp.get_row_value(fba_constr_row)
                else:
                    print(f"FVA min failed for {rid}")
                    flux_ranges[0, idx] = np.nan

                lp.set_obj_dir('maximize')
                res = lp.solve(short_result=True)
                if res['success']:
                    flux_ranges[1, idx] = res['fun']
                    flux_ranges[3, idx] = lp.get_row_value(fba_constr_row)
                else:
                    print(f"FVA max failed for {rid}")
                    flux_ranges[1, idx] = np.nan
        del lp

        res = {'rids': rids, 'min': flux_ranges[0], 'max': flux_ranges[1],
               'obj_min': flux_ranges[2], 'obj_max': flux_ranges[3]}
        return res

    @property
    def obj_dir(self):
        return self.__obj_dir

    @obj_dir.setter
    def obj_dir(self, dir_string):
        if 'min' in dir_string:
            self.__obj_dir = 'minimize'
        else:
            self.__obj_dir = 'maximize'

    def get_flux_bounds(self):
        """retrieve flux bounds for all reactions in a dict.

        :returns: flux bounds configured in the FBA problem
        :rtype: dict (key: reaction id, value: list with two float values for lower/upper bound)
        """
        return {mrid: [bds[0], bds[1]] for mrid, bds in zip(self.mr_rids, self.flux_bounds.T)}

    def set_flux_bounds(self, flux_bounds):
        """selectivly set lower and upper flux bounds for given reactions.

        :param flux_bounds: new lower and upper flux bounds selected reactions
        :type flux_bounds: dict (key: reaction id, value: list with two float values for lower/upper bound)
        """
        for rid, (lb, ub) in flux_bounds.items():
            self.flux_bounds[0, self.mrid2idx[rid]] = lb
            self.flux_bounds[1, self.mrid2idx[rid]] = ub

    def get_obj_coefs(self):
        """retrieve the coefficients of the FBA objective function.

        :returns: objective coefficients
        :rtype: dict (key: reaction id, value: float)
        """
        return {rid: self.obj_coefs[idx] for rid, idx in self.mrid2idx.items()
                if self.obj_coefs[idx] != 0.0}

    def set_obj_coefs(self, coefficiants):
        """set the coefficients of the FBA objective function.

        :param coefficiants: objective coefficients of selected reactions
        :type coefficiants: dict (key: reaction id, value: float)
        """
        self.obj_coefs = np.zeros(len(self.mr_rids))
        for rid, coef in coefficiants.items():
            self.obj_coefs[self.mrid2idx[rid]] = coef
