"""Implementation of GBA Problem class.

   - several process machines supported (yet to be validated)
   - non-enzymatic metabolic reactions supported (yet to be validated)

   Yet to implement:
   - enzyme degradation
   - enzyme transitions
   - several compartments constraints
   - target concentrations and target fluxes

Peter Schubert, HHU Duesseldorf, June 2022
"""

import numpy as np

from xbanalysis.problems.lp_problem import LpProblem


class RbaProblem:

    def __init__(self, xba_model):
        """

        :param xba_model:
        """
        # metabolic reactions and enzyme transitions (no FBA reactions, no protein synthesis, no degradation)
        self.mr_rids = [r.id for r in xba_model.reactions.values() if r.sboterm.is_in_branch('SBO:0000167')
                        and r.sboterm.is_in_branch('SBO:0000179') is False]
        self.mrid2enz = {rid: xba_model.reactions[rid].enzyme for rid in self.mr_rids}
        self.enz_sids = [enz_sid for enz_sid in self.mrid2enz.values() if enz_sid is not None]
        self.rev_enz_sids = [enz_sid for mr_rid, enz_sid in self.mrid2enz.items()
                             if enz_sid is not None and xba_model.reactions[mr_rid].reversible]
        self.processes = {s.rba_process_machine: sid for sid, s in xba_model.species.items()
                          if hasattr(s, 'rba_process_machine')}
        self.psenz_rids = [xba_model.species[enz_sid].ps_rid for enz_sid in self.enz_sids]
        self.pspm_rids = [xba_model.species[enz_sid].ps_rid for enz_sid in self.processes.values()]
        self.densities = {cid: c.rba_frac_gcdw for cid, c in xba_model.compartments.items()
                          if hasattr(c, 'rba_frac_gcdw')}

        # metabolites and enzymes used in metabolic reactions and enzyme transitions
        sids = set()
        for rid in self.mr_rids:
            for sid in xba_model.reactions[rid].reactants:
                sids.add(sid)
            for sid in xba_model.reactions[rid].products:
                sids.add(sid)
        used_sids = list(sids)
        self.sids = [sid for sid in used_sids if xba_model.species[sid].constant is False]
        self.ext_conc = {sid: xba_model.species[sid].initial_conc
                         for sid in used_sids if xba_model.species[sid].constant is True}

        self.rba_cols = self.mr_rids + self.enz_sids + list(self.processes)
        self.rba_rows = (self.sids + list(self.processes)
                         + [sid + '_feff' for sid in self.enz_sids]
                         + [sid + '_reff' for sid in self.rev_enz_sids]
                         + list(self.densities))

        self.is_rba_model = self.check_rba_model(xba_model)
        if self.is_rba_model is False:
            print('Error: not a GBA model - missing parameters')
            return

        fix_m, var_m, row_bnds_m = self.construct_mass_balance(xba_model)
        fix_pc, var_pc, row_bnds_pc = self.construct_process_capacities(xba_model)
        fix_eff, var_eff, row_bnds_eff = self.construct_efficencies(xba_model)
        fix_cd, var_cd, row_bnds_cd = self.construct_densities(xba_model)
        self.fix_cd = fix_cd
        self.fix_mat = np.vstack((fix_m, fix_pc, fix_eff, fix_cd))
        self.var_mat = np.vstack((var_m, var_pc, var_eff, var_cd))
        self.fix_row_bnds = np.vstack((np.zeros_like(row_bnds_m), np.zeros_like(row_bnds_pc),
                                       row_bnds_eff, row_bnds_cd))
        self.var_row_bnds = np.vstack((row_bnds_m, row_bnds_pc,
                                       np.zeros_like(row_bnds_eff), np.zeros_like(row_bnds_cd)))

        # in RBA we only check feasibility.
        # Here we maximize enzyme amount and check that objective value e.g. > 1e-10 mmol/gcdw
        self.obj_dir = 'maximize'
        self.obj_coefs = np.hstack((np.zeros(len(self.mr_rids)),
                                    np.ones(len(self.enz_sids) + len(self.processes))))
        col_lbs = np.hstack(([xba_model.reactions[rid].fbc_lb for rid in self.mr_rids],
                             np.zeros(len(self.enz_sids) + len(self.processes))))
        col_ubs = np.hstack(([xba_model.reactions[rid].fbc_ub for rid in self.mr_rids],
                             np.nan * np.ones(len(self.enz_sids) + len(self.processes))))
        self.col_bnds = np.vstack((col_lbs, col_ubs))
        self.scaling = '2N'
        self.max_iter = 20

    def check_rba_model(self, xba_model):
        """Check if rba related data available in the XBA model

        :param xba_model: Xba model with data extracted from SBML file
        :type xba_model: XbaModel
        :return: indicator if model contains data required for RBA problem formulation
        :rtype: bool
        """
        is_rba_model = True

        r_check = {'flux lower bound': 'fbc_lb', 'flux upper bound': 'fbc_ub', 'kapp forward': 'rba_kapp_f'}
        for check, param in r_check.items():
            missing = [rid for rid in self.mr_rids if hasattr(xba_model.reactions[rid], param) is False]
            if len(missing) > 0:
                print(f'missing {check} in:', missing)
                is_rba_model = False

        missing = [rid for rid in self.mr_rids if xba_model.reactions[rid].reversible
                   and hasattr(xba_model.reactions[rid], 'rba_kapp_r') is False]
        if len(missing) > 0:
            print('missing kapp reverse in:', missing)
            is_rba_model = False

        missing = [sid for sid in self.sids + self.enz_sids + list(self.processes.values())
                   if hasattr(xba_model.species[sid], 'mw') is False]
        if len(missing) > 0:
            print('missing molecular weight in:', missing)
            is_rba_model = False

        missing = [sid for sid in list(self.processes.values())
                   if hasattr(xba_model.species[sid], 'rba_process_capacity') is False]
        if len(missing) > 0:
            print('missing process capacity in:', missing)
            is_rba_model = False

        if len(self.densities) == 0:
            print('missing a density constraint')
            is_rba_model = False
        return is_rba_model

    def rba_optimize(self):
        """

        :return:
        """
        if self.is_rba_model is False:
            print('Error: not a RBA model - missing parameters')
            res = {'success': False, 'fun': 0.0, 'message': 'not an RBA model'}
            return res

        res = {}
        # Preanalysis: find range of growth rates
        gr = {'lb': 0.0, 'ub': np.nan, 'try': 1.0}
        for i in range(self.max_iter):
            constr_coefs = self.fix_mat + gr['try'] * self.var_mat
            # row_bnds = self.fix_row_bnds + gr['try'] * self.var_row_bnds
            lp = LpProblem()
            lp.configure(self.obj_coefs, constr_coefs, self.fix_row_bnds, self.col_bnds, self.obj_dir)
            res = lp.solve(scaling=self.scaling)
            del lp
            if res['success'] and res['fun'] > 1e-10:
                gr['lb'] = gr['try']
                gr['try'] *= 10
            else:
                if gr['try'] <= 1e-4:
                    break
                gr['ub'] = gr['try']
                gr['try'] /= 10
            if gr['lb'] > 0.0 and np.isfinite(gr['ub']):
                break

        if gr['lb'] == 0.0 or np.isnan(gr['ub']):
            res = {'success': False, 'fun': 0.0, 'message': 'no solution found'}
            return res

        # find exact growth rate
        for i in range(self.max_iter):
            gr['try'] = 0.5 * (gr['lb'] + gr['ub'])
            constr_coefs = self.fix_mat + gr['try'] * self.var_mat
            # row_bnds = self.fix_row_bnds + gr['try'] * self.var_row_bnds
            lp = LpProblem()
            lp.configure(self.obj_coefs, constr_coefs, self.fix_row_bnds, self.col_bnds, self.obj_dir)
            res = lp.solve(scaling=self.scaling)
            del lp
            if res['success'] and res['fun'] > 1e-10:
                gr['lb'] = gr['try']
            else:
                gr['ub'] = gr['try']
            if (gr['ub'] - gr['lb']) < gr['try'] * 1e-5:
                break
        res['fun'] = gr['try']
        return res

    def construct_mass_balance(self, xba_model):

        fix_m_x_mr = xba_model.get_stoic_matrix(self.sids, self.mr_rids)
        var_m_x_enz = xba_model.get_stoic_matrix(self.sids, self.psenz_rids)
        var_m_x_pm = xba_model.get_stoic_matrix(self.sids, self.pspm_rids)
        fix_m = np.hstack((fix_m_x_mr,
                           np.zeros((len(self.sids), len(self.enz_sids)+len(self.processes)))))
        var_m = np.hstack((np.zeros((len(self.sids), len(self.mr_rids))),
                           var_m_x_enz,
                           var_m_x_pm))
        targets = np.array([getattr(xba_model.species[sid], 'rba_target', 0.0) for sid in self.sids])
        row_bnds_m = np.vstack((targets, targets)).T
        return fix_m, var_m, row_bnds_m

    def construct_process_capacities(self, xba_model):

        var_pc_x_enz = np.zeros((len(self.processes), len(self.psenz_rids)))
        for idx, pm in enumerate(self.processes):
            var_pc_x_enz[idx] = [xba_model.species[enz_sid].rba_process_costs.get(pm, 0.0)
                                 for enz_sid in self.enz_sids]
        var_pc_x_pm = np.zeros((len(self.processes), len(self.processes)))
        for idx, pm in enumerate(self.processes):
            var_pc_x_pm[idx] = [xba_model.species[enz_sid].rba_process_costs.get(pm, 0.0)
                                for enz_sid in self.processes.values()]
        fix_pc_x_pm = -np.diag([xba_model.species[enz_sid].rba_process_capacity
                                for enz_sid in self.processes.values()]) * 3600
        fix_pc = np.hstack((np.zeros((len(self.processes), len(self.mr_rids) + len(self.enz_sids))),
                            fix_pc_x_pm))
        var_pc = np.hstack((np.zeros((len(self.processes), len(self.mr_rids))),
                            var_pc_x_enz, var_pc_x_pm))
        row_bnds_pc = np.zeros((len(self.processes), 2))
        return fix_pc, var_pc, row_bnds_pc

    def get_efficiencies(self, xba_model):

        efficiencies = np.zeros((2, len(self.enz_sids)))
        idx = 0
        for mr_rid, enz_sid in self.mrid2enz.items():
            if enz_sid is not None:
                r = xba_model.reactions[mr_rid]
                saturation = 1.0
                if hasattr(r, 'rba_km_ext'):
                    for sid in r.reactants:
                        if sid in self.ext_conc:
                            saturation *= self.ext_conc[sid] / (self.ext_conc[sid] + r.rba_km_ext)
                efficiencies[0, idx] = r.rba_kapp_f * 3600 * saturation
                saturation = 1.0
                if r.reversible:
                    if hasattr(r, 'rba_km_ext'):
                        for sid in r.products:
                            if sid in self.ext_conc:
                                saturation *= self.ext_conc[sid] / (self.ext_conc[sid] + r.rba_km_ext)
                    efficiencies[1, idx] = r.rba_kapp_r * 3600 * saturation
                idx += 1
        return efficiencies

    def construct_efficencies(self, xba_model):

        efficiencies = self.get_efficiencies(xba_model)
        mask = np.array([enz_sid is not None for enz_sid in self.mrid2enz.values()])
        fix_feff_x_mr = np.diag(np.ones(len(self.mrid2enz)))[mask]
        fix_feff_x_enz = np.diag(-efficiencies[0])
        fix_feff_x_pm = np.zeros((len(self.enz_sids), len(self.processes)))
        fix_feff = np.hstack((fix_feff_x_mr, fix_feff_x_enz, fix_feff_x_pm))
        row_bnds_feff = np.vstack((np.ones(len(self.enz_sids)) * np.nan, np.zeros(len(self.enz_sids)))).T

        mask = [enz_sid is not None and xba_model.reactions[mr_rid].reversible
                for mr_rid, enz_sid in self.mrid2enz.items()]
        fix_reff_x_mr = np.diag(-np.ones(len(self.mrid2enz)))[mask]
        mask = [enz_sid in self.rev_enz_sids for enz_sid in self.enz_sids]
        fix_reff_x_enz = np.diag(-efficiencies[1])[mask]
        fix_reff_x_pm = np.zeros((len(self.rev_enz_sids), len(self.processes)))
        fix_reff = np.hstack((fix_reff_x_mr, fix_reff_x_enz, fix_reff_x_pm))
        row_bnds_reff = np.vstack((np.ones(len(self.rev_enz_sids)) * np.nan, np.zeros(len(self.rev_enz_sids)))).T
        fix_eff = np.vstack((fix_feff, fix_reff))
        var_eff = np.zeros_like(fix_eff)
        row_bnds_eff = np.vstack((row_bnds_feff, row_bnds_reff))
        return fix_eff, var_eff, row_bnds_eff

    def construct_densities(self, xba_model):
        fix_cd_x_mr = np.zeros((len(self.densities), len(self.mr_rids)))
        fix_cd_x_enz = np.zeros((len(self.densities), len(self.enz_sids)))
        for row_idx, cid in enumerate(self.densities):
            for col_idx, sid in enumerate(self.enz_sids):
                if xba_model.species[sid].compartment == cid:
                    fix_cd_x_enz[row_idx, col_idx] = xba_model.species[sid].mw / 1000
        fix_cd_x_pm = np.zeros((len(self.densities), len(self.processes)))
        for row_idx, cid in enumerate(self.densities):
            for col_idx, sid in enumerate(self.processes.values()):
                if xba_model.species[sid].compartment == cid:
                    fix_cd_x_pm[row_idx, col_idx] = xba_model.species[sid].mw / 1000
        fix_cd = np.hstack((fix_cd_x_mr, fix_cd_x_enz, fix_cd_x_pm))

        var_cd = np.zeros_like(fix_cd)
        row_bnds_cd = np.vstack((np.zeros(len(self.densities)),
                                 [val for val in self.densities.values()])).T
        return fix_cd, var_cd, row_bnds_cd
